import time
from datetime import datetime, timedelta

from pipeline.cloud import http
from pipeline.cloud.schemas.runs import ClusterRunResult, RunState


def get_run(run_id: str):
    response = http.get(f"/v4/runs/{run_id}")
    result = ClusterRunResult.parse_raw(response.text)
    return result


def poll_for_run_completion(run_id, timeout_secs: int = 5 * 60, interval_secs: int = 5):
    start_time = datetime.now()
    while datetime.now() - start_time < timedelta(seconds=timeout_secs):
        run = get_run(run_id)
        if RunState.is_terminal(run.state):
            return run
        print(f"Waiting for run {run_id} to finish...")
        time.sleep(interval_secs)
    raise TimeoutError(f"Run {run_id} did not finish in the allotted time")



from httpx import HTTPStatusError

from pipeline.cloud import http
from pipeline.cloud.schemas import pointers as pointer_schemas


def create_pointer(
    pointer: str,
    target_pipeline_id_or_pointer: str,
    overwrite: bool = False,
    locked: bool = False,
) -> None:
    try:
        response = http.post(
            "/v3/pointers",
            json_data=pointer_schemas.PointerCreate(
                pointer=pointer,
                pointer_or_pipeline_id=target_pipeline_id_or_pointer,
                locked=locked,
            ).dict(),
            handle_error=False,
        )
        response.raise_for_status()
    except HTTPStatusError as e:
        if e.response.status_code == 409 and overwrite:
            http.patch(
                f"/v3/pointers/{pointer}",
                json_data=pointer_schemas.PointerPatch(
                    pointer_or_pipeline_id=target_pipeline_id_or_pointer,
                    locked=locked,
                ).dict(),
            )
        elif e.response.status_code == 404:
            print(e.response.text)
            print(f"Pipeline {target_pipeline_id_or_pointer} not found")
        else:
            raise



import json
import os
import sys
import typing as t
from pathlib import Path

import yaml
from pydantic import BaseModel

PIPELINE_DIR = Path(
    os.getenv(
        "PIPELINE_DIR",
        Path(os.getenv("LOCALAPPDATA")) / ".pipeline/"
        if (sys.platform == "win32" or sys.platform == "cygwin")
        else Path.home() / ".pipeline/",
    )
)


class _RemoteModel(BaseModel):
    alias: t.Optional[str]
    url: str
    token: t.Optional[str]


class _ConfigurationModel(BaseModel):
    active_remote: t.Optional[str]
    remotes: t.Optional[t.List[_RemoteModel]]


class Configuration:
    def __init__(
        self,
    ) -> None:
        self._config: _ConfigurationModel | None = None

        PIPELINE_DIR.mkdir(exist_ok=True)
        (PIPELINE_DIR / "files").mkdir(exist_ok=True)
        self._debug_mode = os.environ.get("PIPELINE_DEBUG", "0") == "1"

    @property
    def active_remote(self) -> _RemoteModel | None:
        if self._config is not None and self._config.remotes is not None:
            if self._config.active_remote is not None:
                for remote in self._config.remotes:
                    if remote.alias == self._config.active_remote:
                        return remote
            else:
                if len(self._config.remotes) > 0:
                    new_remote_alias = self._config.remotes[0].alias
                    if new_remote_alias is None:
                        raise ValueError("Remote alias cannot be None")
                    self.set_active_remote(new_remote_alias)

                    return self._config.remotes[0]

                print("No active remote set, please run 'pipeline cluster use <name>'")
        return None

    def set_debug_mode(self, debug: bool) -> None:
        self._debug_mode = debug

    def is_debugging(self) -> bool:
        return self._debug_mode

    @property
    def files_cache(self) -> Path:
        return PIPELINE_DIR / "files"

    @property
    def remotes(self) -> t.List[_RemoteModel] | None:
        if self._config is None:
            return None
        return self._config.remotes

    def load(self) -> None:
        path = PIPELINE_DIR / "config.yaml"
        if not path.exists():
            self._config = _ConfigurationModel(
                active_remote=None,
                remotes=None,
            )
            return

        try:
            with open(path, "r") as configuration_file:
                self._config = _ConfigurationModel.parse_obj(
                    yaml.load(
                        configuration_file,
                        Loader=yaml.FullLoader,
                    ),
                )
        except Exception:
            raise Exception("Invalid configuration file at '~/.pipeline/config.yaml'")

    def save(self) -> None:
        if self._config is None:
            raise ValueError("No configuration loaded")

        path = PIPELINE_DIR / "config.yaml"

        PIPELINE_DIR.mkdir(exist_ok=True)

        with open(path, "w") as configuration_file:
            yaml.dump(
                json.loads(
                    self._config.json(
                        exclude_none=True,
                    )
                ),
                configuration_file,
            )

    def set_active_remote(self, alias: str) -> None:
        if self._config is None:
            raise ValueError("No configuration loaded")

        if self._config.remotes is None:
            raise ValueError("No remotes configured")

        if not any([remote.alias == alias for remote in self._config.remotes]):
            raise ValueError(f"Remote with alias '{alias}' does not exist")

        self._config.active_remote = alias

        self.save()

    def add_remote(
        self,
        alias: str,
        url: str,
        token: str,
    ) -> None:
        if self._config is None:
            raise ValueError("No configuration loaded")

        if self._config.remotes is None:
            self._config.remotes = []

        if any([remote.alias == alias for remote in self._config.remotes]):
            raise ValueError(f"Remote with alias '{alias}' already exists")

        self._config.remotes.append(
            _RemoteModel(
                alias=alias,
                url=url,
                token=token,
            )
        )

        if self._config.active_remote is None:
            self._config.active_remote = alias

        self.save()

    def remove_remote(self, alias: str) -> None:
        if self._config is None:
            raise ValueError("No configuration loaded")

        if self._config.remotes is None:
            return

        alias_index_array = [
            idx for idx, obj in enumerate(self._config.remotes) if obj.alias == alias
        ]
        if len(alias_index_array) == 0:
            raise ValueError(f"Remote with alias '{alias}' does not exist")

        self._config.remotes.pop(alias_index_array[0])


current_configuration = Configuration()
current_configuration.load()



from enum import Enum


class Accelerator(str, Enum):
    nvidia_t4: str = "nvidia_t4"
    nvidia_a100: str = "nvidia_a100"
    nvidia_a100_80gb: str = "nvidia_a100_80gb"
    nvidia_h100: str = "nvidia_h100"
    nvidia_l4: str = "nvidia_l4"

    cpu: str = "cpu"

    nvidia_a100_5gb: str = "nvidia_a100_5gb"
    nvidia_a100_10gb: str = "nvidia_a100_10gb"
    nvidia_a100_20gb: str = "nvidia_a100_20gb"

    nvidia_a100_80gb_10gb: str = "nvidia_a100_80gb_10gb"
    nvidia_a100_80gb_20gb: str = "nvidia_a100_80gb_20gb"
    nvidia_a100_80gb_40gb: str = "nvidia_a100_80gb_40gb"

    nvidia_a10: str = "nvidia_a10"
    nvidia_a10_12gb: str = "nvidia_a10_12gb"
    nvidia_a10_8gb: str = "nvidia_a10_8gb"
    nvidia_a10_4gb: str = "nvidia_a10_4gb"

    @classmethod
    def from_str(cls, accelerator: str) -> "Accelerator":
        if "T4" in accelerator:
            accelerator_type = Accelerator.nvidia_t4
        elif "A100" in accelerator:
            if "80GB" in accelerator:
                accelerator_type = Accelerator.nvidia_a100_80gb
            else:
                accelerator_type = Accelerator.nvidia_a100
        elif "H100" in accelerator:
            accelerator_type = Accelerator.nvidia_h100
        elif "L4" in accelerator:
            accelerator_type = Accelerator.nvidia_l4
        # guess this works as long as its after A100
        elif "A10" in accelerator:
            accelerator_type = Accelerator.nvidia_a10

        else:
            raise Exception(f"Unknown GPU name: {accelerator}")

        return accelerator_type

    @classmethod
    def valid_accelerator_config(cls, accelerators: list["Accelerator"]):
        return accelerators in [
            [Accelerator.nvidia_t4],
            [Accelerator.nvidia_a100],
            [Accelerator.nvidia_a100] * 2,
            [Accelerator.nvidia_a100] * 4,
            [Accelerator.nvidia_a100_80gb],
            [Accelerator.nvidia_a100_80gb] * 2,
            [Accelerator.nvidia_a100_80gb] * 4,
            [Accelerator.nvidia_l4],
            [Accelerator.nvidia_a10],
            [Accelerator.cpu],
        ]


nvidia_gpus = [
    Accelerator.nvidia_t4,
    Accelerator.nvidia_a100,
    Accelerator.nvidia_a100_80gb,
    Accelerator.nvidia_h100,
    Accelerator.nvidia_l4,
    Accelerator.nvidia_a100_5gb,
    Accelerator.nvidia_a100_10gb,
    Accelerator.nvidia_a100_20gb,
    Accelerator.nvidia_a100_80gb_10gb,
    Accelerator.nvidia_a100_80gb_20gb,
    Accelerator.nvidia_a100_80gb_40gb,
    Accelerator.nvidia_a10,
    Accelerator.nvidia_a10_12gb,
    Accelerator.nvidia_a10_8gb,
    Accelerator.nvidia_a10_4gb,
]



import urllib.parse

import httpx


def authenticate(token: str, url: str) -> bool:
    auth_url = urllib.parse.urljoin(url, "/v4/tokens/validate")
    response = httpx.get(
        auth_url,
        headers={
            "Authorization": f"Bearer {token}",
        },
    )
    response.raise_for_status()



import typing as t
from enum import Enum
from math import ceil

from pydantic import conint

from . import BaseModel, GenericModel

DataType = t.TypeVar("DataType")

PAGINATION_LIMIT = 10_000
DEFAULT_LIMIT = 1000


class Pagination(BaseModel):
    """Query parameters for requesting a set of items of a given resource,
    in paginated form."""

    #: Index of the first item to return within the total queryset
    skip: conint(ge=0)  # type: ignore

    #: Maximum number of items to return
    limit: conint(ge=1, le=PAGINATION_LIMIT)  # type: ignore


class Paginated(GenericModel, t.Generic[DataType]):
    """Response model for a paginated set of a given resource."""

    #: Index of the first item to return within the total queryset
    skip: conint(ge=0)  # type: ignore

    #: Maximum number of items the `data` field will contain
    limit: conint(ge=1, le=PAGINATION_LIMIT)  # type: ignore

    #: Number of items within the total queryset
    total: conint(ge=0)  # type: ignore

    #: Resource data, which should be immutable
    data: t.Sequence[DataType]

    @classmethod
    def of(cls, data: t.Sequence[DataType], pagination: Pagination, total: int):
        return Paginated(**pagination.dict(), total=total, data=data)


def get_default_pagination():
    return Pagination(skip=0, limit=DEFAULT_LIMIT)


class PagePosition(t.TypedDict):
    current: int
    total: int


def to_page_position(skip: int, items_per_page: int, total_items: int) -> PagePosition:
    """Converts cursor position into page position"""
    current_page = skip // items_per_page + 1
    total_pages = max(1, ceil(total_items / items_per_page))
    return PagePosition(current=current_page, total=total_pages)


class Order(str, Enum):
    descending = "descending"
    ascending = "ascending"



import json
import os
import subprocess
import sys
import typing as t
from argparse import Namespace
from pathlib import Path

import docker
import yaml
from docker.types import DeviceRequest, LogConfig
from pydantic import BaseModel

from pipeline.cloud import http
from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.schemas import cluster as cluster_schemas
from pipeline.cloud.schemas import pipelines as pipelines_schemas
from pipeline.cloud.schemas import pointers as pointers_schemas
from pipeline.cloud.schemas import registry as registry_schemas
from pipeline.container import docker_templates
from pipeline.util.logging import _print


class PythonRuntime(BaseModel):
    version: str
    requirements: t.List[str] | None

    class Config:
        extra = "forbid"


class RuntimeConfig(BaseModel):
    container_commands: t.List[str] | None
    python: PythonRuntime | None

    class Config:
        extra = "forbid"


class PipelineConfig(BaseModel):
    runtime: RuntimeConfig
    accelerators: t.List[Accelerator] = []
    accelerator_memory: int | None
    pipeline_graph: str
    pipeline_name: str = ""
    description: str | None = None
    readme: str | None = None
    extras: t.Dict[str, t.Any] | None
    cluster: cluster_schemas.PipelineClusterConfig | None = None

    class Config:
        extra = "forbid"


def _edit_pointer(
    existing_pointer: str,
    pointer_or_pipeline_id: str,
):
    edit_schema = pointers_schemas.PointerPatch(
        pointer_or_pipeline_id=pointer_or_pipeline_id,
        locked=False,
    )

    result = http.patch(
        f"/v4/pointers/{existing_pointer}",
        json.loads(
            edit_schema.json(),
        ),
    )

    if result.status_code == 200:
        pointer = pointers_schemas.PointerGet.parse_obj(result.json())
        _print(f"Updated pointer {pointer.pointer} -> {pointer.pipeline_id}")
    else:
        _print(f"Failed to edit pointer {existing_pointer}", "ERROR")


def _create_pointer(
    new_pointer: str,
    pointer_or_pipeline_id: str,
    force=False,
) -> None:
    create_schema = pointers_schemas.PointerCreate(
        pointer=new_pointer,
        pointer_or_pipeline_id=pointer_or_pipeline_id,
        locked=False,
    )
    result = http.post(
        "/v4/pointers",
        json.loads(
            create_schema.json(),
        ),
        handle_error=False,
    )

    if result.status_code == 409:
        if force:
            _print("Pointer already exists, forcing update", "WARNING")
            _edit_pointer(new_pointer, pointer_or_pipeline_id)
        else:
            _print(
                f"Pointer {new_pointer} already exists, use --pointer-overwrite to update",  # noqa
                "WARNING",
            )
        return
    elif result.status_code == 201:
        pointer = pointers_schemas.PointerGet.parse_obj(result.json())
        _print(f"Created pointer {pointer.pointer} -> {pointer.pipeline_id}")
    else:
        raise ValueError(f"Failed to create pointer {new_pointer}\n{result.text}")


def _up_container(namespace: Namespace):
    _print("Starting container...", "INFO")
    config_file = Path("./pipeline.yaml")

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    pipeline_name = pipeline_config.pipeline_name
    docker_client = docker.from_env()
    lc = LogConfig(
        type=LogConfig.types.JSON,
        config={
            "max-size": "1g",
        },
    )
    volumes: list | None = None

    port = int(getattr(namespace, "port", "14300"))

    run_command = [
        "uvicorn",
        "pipeline.container.startup:create_app",
        "--host",
        "0.0.0.0",
        "--port",
        str(port),
        "--factory",
    ]

    environment_variables = dict()

    if getattr(namespace, "debug", False):
        run_command.append("--reload")
        current_path = Path("./").expanduser().resolve()
        if volumes is None:
            volumes = []

        volumes.append(f"{current_path}:/app/")
        environment_variables["DEBUG"] = "1"
        environment_variables["LOG_LEVEL"] = "DEBUG"
        environment_variables["FASTAPI_ENV"] = "development"

    if extra_volumes := getattr(namespace, "volume", None):
        if volumes is None:
            volumes = []

        for volume in extra_volumes:
            if ":" not in volume:
                raise ValueError(
                    f"Invalid volume {volume}, must be in format host_path:container_path"  # noqa
                )

            local_path = Path(volume.split(":")[0]).expanduser().resolve()
            container_path = Path(volume.split(":")[1])

            volumes.append(f"{local_path}:{container_path}")

    gpu_ids: list | None = None
    try:
        gpu_ids = [
            f"{i}"
            for i in range(
                0,
                len(
                    subprocess.check_output(
                        [
                            "nvidia-smi",
                            "-L",
                        ]
                    )
                    .decode()
                    .splitlines()
                ),
            )
        ]
    except Exception:
        gpu_ids = None

    # Stop container on python exit
    container = docker_client.containers.run(
        pipeline_name,
        ports={f"{port}/tcp": int(port)},
        stderr=True,
        stdout=True,
        log_config=lc,
        remove=True,
        auto_remove=True,
        detach=True,
        device_requests=(
            [DeviceRequest(device_ids=gpu_ids, capabilities=[["gpu"]])]
            if gpu_ids
            else None
        ),
        command=run_command,
        volumes=volumes,
        environment=environment_variables,
    )

    _print(
        f"Container started on port {port}.\n\n\t\tView the live docs:\n\n\t\t\t http://localhost:{port}/redoc\n\n\t\tor live play:\n\n\t\t\t http://localhost:{port}/play\n",  # noqa
        "SUCCESS",
    )

    while True:
        try:
            for line in container.logs(stream=True):
                print(line.decode("utf-8").strip())
        except KeyboardInterrupt:
            _print("Stopping container...", "WARNING")
            container.stop()
            # container.remove()
            break
        except docker.errors.NotFound:
            _print("Container did not start successfully", "ERROR")
            break


def _build_container(namespace: Namespace):
    _print("Starting build service...", "INFO")
    template = docker_templates.dockerfile_template

    config_file = Path("./pipeline.yaml")

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    if not pipeline_config.runtime:
        raise ValueError("No runtime config found")
    if not pipeline_config.runtime.python:
        raise ValueError("No python runtime config found")

    python_runtime = pipeline_config.runtime.python
    dockerfile_str = template.format(
        python_version=python_runtime.version,
        python_requirements=(
            " ".join(python_runtime.requirements) if python_runtime.requirements else ""
        ),
        container_commands="".join(
            [
                "RUN " + command + " \n"
                for command in pipeline_config.runtime.container_commands or []
            ]
        ),
        pipeline_path=pipeline_config.pipeline_graph,
        pipeline_name=pipeline_config.pipeline_name,
        pipeline_image=pipeline_config.pipeline_name,
    )

    dockerfile_path = Path("./pipeline.dockerfile")
    dockerfile_path.write_text(dockerfile_str)
    docker_client = docker.APIClient()
    generator = docker_client.build(
        # fileobj=dockerfile_path.open("rb"),
        path="./",
        # custom_context=True,
        dockerfile=dockerfile_path.absolute(),
        # tag="test",
        rm=True,
        decode=True,
        platform="linux/amd64",
    )
    docker_image_id = None
    while True:
        try:
            output = generator.__next__()
            if "aux" in output:
                docker_image_id = output["aux"]["ID"]
            if "stream" in output:
                _print(output["stream"].strip("\n"))
            if "errorDetail" in output:
                raise Exception(output["errorDetail"])
        except StopIteration:
            _print("Docker image build complete.")
            break

    docker_client = docker.from_env()
    new_container = docker_client.images.get(docker_image_id)

    created_image_full_id = new_container.id.split(":")[1]
    created_image_short_id = created_image_full_id[:12]

    _print(f"Built container {created_image_short_id}", "SUCCESS")

    pipeline_repo = (
        pipeline_config.pipeline_name.split(":")[0]
        if ":" in pipeline_config.pipeline_name
        else pipeline_config.pipeline_name
    )
    # pipeline_tag = (
    #     pipeline_config.pipeline_name.split(":")[1]
    #     if ":" in pipeline_config.pipeline_name
    #     else None
    # )

    new_container.tag(pipeline_repo)
    _print(f"Created tag {pipeline_repo}", "SUCCESS")

    new_container.tag(pipeline_repo, tag=created_image_short_id)
    _print(f"Created tag {pipeline_repo}:{created_image_short_id}", "SUCCESS")

    # if pipeline_tag:
    #     new_container.tag(pipeline_repo, tag=pipeline_tag)
    #     _print(f"Created tag {pipeline_repo}:{pipeline_tag}", "SUCCESS")


def _push_container(namespace: Namespace):
    """

    Upload protocol:
    1. Request upload URL from server, along with auth token
    2. Upload to URL with auth token
    3. Send complete request to server

    """

    config_file = Path("./pipeline.yaml")

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pointers: list | None = getattr(namespace, "pointer", None)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    # Check for file, transform to string, and put it back in config
    if pipeline_config.readme is not None:
        if os.path.isfile(pipeline_config.readme):
            markdown_file = Path(pipeline_config.readme)
            pipeline_config.readme = markdown_file.read_text()
    else:
        pipeline_config.readme = ""

    # Attempt to format the readme
    readmeless_config = pipeline_config.copy()
    readmeless_config.readme = None
    pipeline_yaml_text = yaml.dump(
        json.loads(readmeless_config.json(exclude_none=True, exclude_unset=True)),
        sort_keys=False,
    )

    pipeline_yaml_text = "```yaml\n" + pipeline_yaml_text + "\n```"
    pipeline_code = Path(
        pipeline_config.pipeline_graph.split(":")[0] + ".py"
    ).read_text()
    pipeline_code = "```python\n" + pipeline_code + "\n```"
    pipeline_config.readme = pipeline_config.readme.format(
        pipeline_name=pipeline_config.pipeline_name,
        pipeline_yaml=pipeline_yaml_text,
        pipeline_code=pipeline_code,
    )

    pipeline_name = (
        pipeline_config.pipeline_name.split(":")[0]
        if ":" in pipeline_config.pipeline_name
        else pipeline_config.pipeline_name
    )

    docker_client = docker.from_env(timeout=300)

    registry_info = http.get(endpoint="/v4/registry")
    registry_info = registry_schemas.RegistryInformation.parse_raw(registry_info.text)

    upload_registry = registry_info.url

    if upload_registry is None:
        raise ValueError("No upload registry found")
    image = docker_client.images.get(pipeline_name)
    image_hash = image.id.split(":")[1]

    hash_tag = image_hash[:12]
    image_to_push = pipeline_name + ":" + hash_tag
    image_to_push_reg = upload_registry + "/" + image_to_push

    upload_token = None
    true_pipeline_name = None
    if registry_info.special_auth:
        start_upload_response = http.post(
            endpoint="/v4/registry/start-upload",
            json_data={
                "pipeline_name": pipeline_name,
                "pipeline_tag": None,
            },
        )
        start_upload_dict = start_upload_response.json()
        upload_token = start_upload_dict.get("bearer", None)
        true_pipeline_name = start_upload_dict.get("pipeline_name")

        if upload_token is None:
            raise ValueError("No upload token found")

        # Login to upload registry
        docker_client.login(
            username="pipeline",
            password=upload_token,
            registry="http://" + upload_registry,
        )

        # Override the tag with the pipeline name from catalyst
        image_to_push = true_pipeline_name + ":" + hash_tag
        image_to_push_reg = upload_registry + "/" + image_to_push

    _print(f"Pushing image to upload registry {upload_registry}", "INFO")

    docker_client.images.get(pipeline_name).tag(image_to_push_reg)
    # Do this after tagging, because we need to use
    # the old pipeline name to tag the local image
    if true_pipeline_name:
        pipeline_name = true_pipeline_name

    resp = docker_client.images.push(
        image_to_push_reg,
        stream=True,
        decode=True,
        auth_config=(
            dict(username="pipeline", password=upload_token) if upload_token else None
        ),
    )

    all_ids = []

    current_index = 0

    for line in resp:
        if "error" in line:
            if line["error"] == "unauthorized: authentication required":
                print(
                    """
Failed to authenticate with the registry.
This can happen if your pipeline took longer than an hour to push.
Please try reduce the size of your pipeline or contact mystic.ai"""
                )
                return
            raise ValueError(line["error"])
        elif "status" in line:
            if "id" not in line or line["status"] != "Pushing":
                continue

            if "id" in line and line["id"] not in all_ids:
                all_ids.append(line["id"])
                print("Adding")

            index_difference = all_ids.index(line["id"]) - current_index

            print_string = (
                line["id"] + " " + line["progress"].replace("\n", "").replace("\r", "")
            )

            if index_difference > 0:
                print_string += "\n" * index_difference + "\r"
                # print("up")
            elif index_difference < 0:
                print_string += "\033[A" * abs(index_difference) + "\r"
                # print("down")
            else:
                print_string += "\r"
                # print("same")
            current_index = all_ids.index(line["id"])

            sys.stdout.write(print_string)
            sys.stdout.flush()

    new_deployment_request = http.post(
        endpoint="/v4/pipelines",
        json_data=json.loads(
            pipelines_schemas.PipelineCreate(
                name=pipeline_name,
                image=image_to_push_reg,
                input_variables=[],
                output_variables=[],
                minimum_cache_number=None,
                maximum_cache_number=None,
                gpu_memory_min=pipeline_config.accelerator_memory,
                accelerators=pipeline_config.accelerators,
                description=pipeline_config.description,
                readme=pipeline_config.readme,
                extras=pipeline_config.extras,
                cluster=pipeline_config.cluster,
            ).json()
        ),
    )

    new_deployment = pipelines_schemas.PipelineGet.parse_obj(
        new_deployment_request.json()
    )

    _print(
        f"Created new pipeline deployment for {new_deployment.name} -> {new_deployment.id} (image={new_deployment.image})",  # noqa
        "SUCCESS",
    )

    if pointers:
        pointer_overwrite = getattr(namespace, "pointer_overwrite", False)
        for pointer in pointers:
            _create_pointer(pointer, new_deployment.id, force=pointer_overwrite)


def _init_dir(namespace: Namespace) -> None:
    _print("Initializing directory...", "INFO")

    pipeline_name = getattr(namespace, "name", None)

    if not pipeline_name:
        pipeline_name = input("Enter a name for your pipeline: ")

    python_template = docker_templates.pipeline_template_python

    default_config = PipelineConfig(
        runtime=RuntimeConfig(
            container_commands=[
                "apt-get update",
                "apt-get install -y git",
            ],
            python=PythonRuntime(
                version="3.10",
                requirements=[
                    "pipeline-ai",
                ],
            ),
        ),
        accelerators=[],
        pipeline_graph="new_pipeline:my_new_pipeline",
        pipeline_name=pipeline_name,
        accelerator_memory=None,
        extras={},
        readme="README.md",
    )
    with open("./pipeline.yaml", "w") as f:
        f.write(yaml.dump(json.loads(default_config.json()), sort_keys=False))

    with open("./new_pipeline.py", "w") as f:
        f.write(python_template)

    with open("./README.md", "w") as f:
        f.write(docker_templates.readme_template)

    _print("Initialized directory.", "SUCCESS")



from argparse import ArgumentParser, _SubParsersAction

from pipeline.console import cluster, container, logs
from pipeline.console.targets import (
    environments,
    files,
    pipelines,
    pointers,
    resources,
    scaling_configs,
)


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    create_parser = command_parser.add_parser(
        "create",
        description="Create a new object.",
        help="Create a new object.",
    )
    create_parser.set_defaults(func=lambda _: create_parser.print_help())

    create_sub_parser = create_parser.add_subparsers(
        dest="target",
    )

    environments.create_parser(create_sub_parser)
    pointers.create_parser(create_sub_parser)
    resources.create_parser(create_sub_parser)
    files.create_parser(create_sub_parser)
    scaling_configs.create_parser(create_sub_parser)


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    edit_parser = command_parser.add_parser(
        "edit",
        description="Edit an object.",
        help="Edit and object.",
    )
    edit_parser.set_defaults(func=lambda _: edit_parser.print_help())

    edit_sub_parser = edit_parser.add_subparsers(
        dest="target",
    )

    environments.edit_parser(edit_sub_parser)
    pipelines.edit_parser(edit_sub_parser)
    pointers.edit_parser(edit_sub_parser)
    files.edit_parser(edit_sub_parser)
    scaling_configs.edit_parser(edit_sub_parser)


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "get",
        description="Get an object(s).",
        help="Get an object(s).",
    )
    get_parser.set_defaults(func=lambda _: get_parser.print_help())

    get_sub_parser = get_parser.add_subparsers(
        dest="target",
    )

    environments.get_parser(get_sub_parser)
    pipelines.get_parser(get_sub_parser)
    pointers.get_parser(get_sub_parser)
    resources.get_parser(get_sub_parser)
    files.get_parser(get_sub_parser)
    scaling_configs.get_parser(get_sub_parser)


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    delete_parser = command_parser.add_parser(
        "delete",
        aliases=["del", "rm"],
        description="Delete an object.",
        help="Delete an object.",
    )
    delete_parser.set_defaults(func=lambda _: delete_parser.print_help())

    delete_sub_parser = delete_parser.add_subparsers(
        dest="target",
    )

    environments.delete_parser(delete_sub_parser)
    pipelines.delete_parser(delete_sub_parser)
    pointers.delete_parser(delete_sub_parser)
    resources.delete_parser(delete_sub_parser)
    files.delete_parser(delete_sub_parser)
    scaling_configs.delete_parser(delete_sub_parser)


def cluster_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    cluster_parser = command_parser.add_parser(
        "cluster",
        aliases=["cl"],
        description="Manage compute clusters.",
        help="Manage compute clusters.",
    )
    cluster_parser.set_defaults(func=lambda _: cluster_parser.print_help())

    cluster_sub_parser = cluster_parser.add_subparsers(
        dest="target",
    )

    cluster.login_parser(cluster_sub_parser)
    cluster.use_parser(cluster_sub_parser)
    cluster.remove_parser(cluster_sub_parser)
    cluster.get_parser(cluster_sub_parser)


def logs_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    logs_parser = command_parser.add_parser(
        "logs",
        description="Get logs.",
        help="Get logs.",
    )
    logs_parser.set_defaults(func=lambda _: logs_parser.print_help())
    logs_sub_parser = logs_parser.add_subparsers(
        dest="target",
    )
    logs.run_logs_parser(logs_sub_parser)
    logs.pipeline_startup_logs_parser(logs_sub_parser)


def container_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    container_parser = command_parser.add_parser(
        "container",
        description="Manage pipeline containers.",
        help="Manage pipeline containers.",
    )
    container_parser.set_defaults(func=lambda _: container_parser.print_help())
    container_sub_parser = container_parser.add_subparsers(
        dest="target",
    )

    build_parser = container_sub_parser.add_parser(
        "build",
        description="Build a pipeline container.",
        help="Build a pipeline container.",
    )
    build_parser.set_defaults(func=container._build_container)

    push_parser = container_sub_parser.add_parser(
        "push",
        description="Push a pipeline container.",
        help="Push a pipeline container.",
    )
    push_parser.set_defaults(func=container._push_container)
    push_parser.add_argument(
        "--pointer",
        "-p",
        action="append",
        help="Pointer for the container.",
    )
    push_parser.add_argument(
        "--pointer-overwrite",
        "-o",
        action="store_true",
        help="Overwrite existing pointers.",
    )

    up_parser = container_sub_parser.add_parser(
        "up",
        description="Start a pipeline container.",
        help="Start a pipeline container.",
    )
    up_parser.set_defaults(func=container._up_container)
    up_parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="Start the container in debug mode.",
    )
    # Allow multiple volumes to be specified

    up_parser.add_argument(
        "--volume",
        "-v",
        action="append",
        help="Mount a volume into the container.",
    )

    up_parser.add_argument(
        "--port",
        "-p",
        help="Container port.",
        default=14300,
    )
    # Init
    init_parser = container_sub_parser.add_parser(
        "init",
        description="Initialize a directory for a new pipeline.",
        help="Initialize a directory for a new pipeline.",
    )

    init_parser.add_argument(
        "--name",
        "-n",
        type=str,
        help="Name of the pipeline.",
    )
    init_parser.set_defaults(func=container._init_dir)



from fastapi import APIRouter
from starlette.responses import RedirectResponse

from pipeline.container.routes.play import router as play_router
from pipeline.container.routes.v4 import router as v4_router

router = APIRouter()
router.include_router(v4_router)
router.include_router(play_router)


@router.get("/")
def redirect_to_play():
    return RedirectResponse("/play")



from argparse import ArgumentParser, Namespace, _SubParsersAction

from pipeline.cloud.logs import get_pipeline_startup_logs, get_run_logs


def run_logs_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    run_logs_parser = command_parser.add_parser("run", help="Get logs for a run.")
    run_logs_parser.set_defaults(func=_run_logs)

    # not supported currently
    # run_logs_parser.add_argument(
    #     "-f",
    #     "--follow",
    #     action="store_true",
    # )

    run_logs_parser.add_argument(
        "run_id",
        help="Run ID to get logs for.",
        type=str,
    )


def _run_logs(args: Namespace) -> None:
    # follow = getattr(args, "follow", False)
    run_id = getattr(args, "run_id")

    log_entries = get_run_logs(run_id)
    if not log_entries:
        return
    for message in log_entries:
        print(message)


def pipeline_startup_logs_parser(
    command_parser: "_SubParsersAction[ArgumentParser]",
) -> None:
    startup_logs_parser = command_parser.add_parser(
        "startup", help="Get logs for a pipeline during startup."
    )
    startup_logs_parser.set_defaults(func=_pipeline_startup_logs)

    # not supported currently
    # startup_logs_parser.add_argument(
    #     "-f",
    #     "--follow",
    #     action="store_true",
    # )

    startup_logs_parser.add_argument(
        "pipeline_id_or_pointer",
        help="Pipeline ID or pointer to get logs for.",
        type=str,
    )


def _pipeline_startup_logs(args: Namespace) -> None:
    # follow = getattr(args, "follow", False)
    pipeline_id_or_pointer = getattr(args, "pipeline_id_or_pointer")

    log_entries = get_pipeline_startup_logs(pipeline_id_or_pointer)
    if not log_entries:
        print("No logs found in the last 24h")
        return
    for message in log_entries:
        print(message)



import os
from datetime import datetime, timezone

import humps
from pydantic import BaseModel as _BaseModel
from pydantic import Extra, validator
from pydantic.generics import GenericModel as _GenericModel

CAMEL_CASE_ALIASES = bool(os.environ.get("CAMEL_CASE_ALIASES", False))


def _generate_alias(s):
    return humps.camelize(s) if CAMEL_CASE_ALIASES else s


class BaseConfig:
    alias_generator = _generate_alias
    allow_population_by_field_name = True
    # Encode datetime objects to JSON as integer timestamps
    json_encoders = {datetime: datetime.timestamp}
    orm_mode = True


class BaseModel(_BaseModel):
    """Base model for schemas."""

    class Config(BaseConfig):
        pass

    @validator("*", pre=True)
    def convert_to_utc(cls, value):
        if isinstance(value, datetime):
            return value.astimezone(timezone.utc)
        return value


class GenericModel(_GenericModel):
    """Base model for schemas with generic typing."""

    class Config(BaseConfig):
        pass


class Patchable(BaseModel):
    """Tag type for marking schemas for patching.

    A Patchable schema is one used for updating other schemas.
    """

    class Config:
        # Forbid fields not specified in the model; this prevents silently
        # failing updates to patch non-matching fields
        extra = Extra.forbid



import io
import os
import typing as t

from pydantic import ValidationError

from pipeline.cloud import http
from pipeline.cloud.files import resolve_run_input_file_object
from pipeline.cloud.schemas.runs import (
    ClusterRunResult,
    RunCreate,
    RunInput,
    RunIOType,
    RunState,
)
from pipeline.objects import File
from pipeline.objects.graph import InputSchema
from pipeline.util.logging import _print
from pipeline.util.streaming import handle_stream_response


class NoResourcesAvailable(Exception):
    """Exception raised when there are no available resources to route the run
    request to.
    """

    def __init__(
        self,
        run_result: ClusterRunResult,
        message=(
            "This pipeline is currently starting up. This is normal behaviour "
            "for pipelines that are new or have not been run in a while. Please "
            "wait a few minutes before next run."
        ),
    ):
        self.message = message
        self.run_result = run_result
        super().__init__(self.message, self.run_result)

    def __str__(self):
        return self.message


def _data_to_run_input(data: t.Tuple) -> t.List[RunInput]:
    input_array = []

    for item in data:
        if isinstance(item, io.IOBase):
            path = os.path.abspath(item.name)
            item = File(path=path)
        input_type = RunIOType.from_object(item)
        if input_type == RunIOType.file or isinstance(item, File):
            input_schema = resolve_run_input_file_object(item)
            input_array.append(input_schema)
            continue
        elif isinstance(item, InputSchema):
            item_dict = item.to_dict()
            output_dict = dict()
            output_dict.update(item_dict)
            for pair_key, pair_value in item_dict.items():
                if isinstance(pair_value, io.IOBase):
                    path = os.path.abspath(pair_value.name)
                    pair_value = File(path=path)
                pair_value_type = RunIOType.from_object(pair_value)
                if pair_value_type == RunIOType.file:
                    new_schema = resolve_run_input_file_object(pair_value)
                    output_dict[pair_key] = new_schema.dict()
                elif pair_value_type == RunIOType.pkl:
                    raise Exception("Generic python objects are not supported yet.")

            input_schema = RunInput(
                type=RunIOType.dictionary,
                value=output_dict,
                file_path=None,
                file_name=None,
                file_url=None,
            )
            input_array.append(input_schema)

            continue
        elif input_type == RunIOType.pkl:
            raise NotImplementedError("Python object input not yet supported")

        input_schema = RunInput(
            type=input_type,
            value=item,
            file_path=None,
            file_name=None,
            file_url=None,
        )
        input_array.append(input_schema)

    return input_array


def _run_pipeline(run_create_schema: RunCreate):
    res = http.post(
        "/v4/runs",
        json_data=run_create_schema.dict(),
        handle_error=False,
    )
    try:
        result = ClusterRunResult.parse_raw(res.text)
    except ValidationError:
        http.raise_if_http_status_error(res)
        raise

    if result.state == RunState.no_resources_available:
        error = NoResourcesAvailable(run_result=result)
        _print(
            f"{error.message}\nRun result:\n{error.run_result.json(indent=2)}",
            level="ERROR",
        )
        raise error
    return result


def run_pipeline(
    pipeline: str,
    *data,
    async_run: bool = False,
    wait_for_resources: bool | None = None,
):
    run_create_schema = RunCreate(
        pipeline=pipeline,
        inputs=_data_to_run_input(data),
        async_run=async_run,
        wait_for_resources=wait_for_resources,
    )

    run_get = _run_pipeline(run_create_schema)
    return run_get


def _stream_pipeline(
    run_create_schema: RunCreate,
) -> t.Generator[ClusterRunResult, t.Any, None]:
    with http.stream(
        method="POST",
        endpoint="/v4/runs/stream",
        json_data=run_create_schema.dict(),
        handle_error=False,
    ) as response:
        for result_json in handle_stream_response(response):
            try:
                result = ClusterRunResult.parse_obj(result_json)
            except ValidationError:
                _print(
                    f"Unexpected result from streaming run:\n"
                    f"Status code = {response.status_code}\n{result_json}"
                )
                return
            except Exception:
                http.raise_if_http_status_error(response)
                raise

            if result.state == RunState.no_resources_available:
                error = NoResourcesAvailable(run_result=result)
                _print(
                    f"{error.message}\nRun result:\n{result.json(indent=2)}",
                    level="ERROR",
                )
                raise error
            yield result


def stream_pipeline(pipeline: str, *data, wait_for_resources: bool | None = None):
    run_create_schema = RunCreate(
        pipeline=pipeline,
        inputs=_data_to_run_input(data),
        wait_for_resources=wait_for_resources,
    )

    for result in _stream_pipeline(run_create_schema):
        yield result



import re
import typing as t

from pydantic import validator

from pipeline.cloud.schemas import BaseModel

POINTER_REGEX = re.compile(
    r"^[a-zA-Z0-9][a-zA-Z0-9-._/]*[a-zA-Z0-9]:[0-9A-Za-z][0-9A-Za-z-_.]{0,127}[0-9A-Za-z]$"  # noqa
)


class PointerGet(BaseModel):
    id: str
    pointer: str
    pipeline_id: str
    locked: bool


class PointerCreate(BaseModel):
    pointer_or_pipeline_id: str
    pointer: str
    locked: t.Optional[bool]

    @validator("pointer")
    def validate_pointer(cls, v):
        if POINTER_REGEX.match(v) is None:
            raise ValueError("Invalid pointer name")
        return v


class PointerPatch(BaseModel):
    pointer_or_pipeline_id: t.Optional[str]
    locked: t.Optional[bool]



import json
from argparse import ArgumentParser, _SubParsersAction

from tabulate import tabulate

from pipeline.cloud import http
from pipeline.cloud.compute_requirements import Accelerator

# Parser builder


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "resources", help="Get resource information."
    )
    get_parser.set_defaults(func=lambda _: list_resources())


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


# Functions


def _shorten_id(id: str, leading_length: int = 4, trailing_length: int = 4) -> str:
    if id == "None":
        return "None"
    return id[:leading_length] + "..." + id[-trailing_length:]


def list_resources() -> None:
    resource_information = http.get("/v3/core/resources").json()
    resource_information = [json.loads(resource) for resource in resource_information]

    resource_data = [
        [
            _shorten_id(resource["id"]),
            [
                _shorten_id(p_id)
                for cached_pipelines in resource["pipeline_cache"].values()
                for p_id in cached_pipelines
            ],
            _shorten_id(str(resource["current_run"]))
            if (resource["busy"] == 1 and resource["current_run"] != -1)
            else "-",
            [_shorten_id(id) for id in resource["run_queue"]],
            "cpu"
            if not (accelerators := resource.get("gpus", None))
            else (
                "cpu"
                if "cpu" in accelerators
                else "\n".join(
                    [
                        f"{[accel['name'] for accel in accelerators].count(accelerator)}× {Accelerator.from_str(accelerator)} ({round(sum([accel['vram_total_mb'] for accel in accelerators if accel['name'] == accelerator]) / 1024.0, 1)}GB VRAM)"  # noqa E501
                        for accelerator in set(
                            [accel["name"] for accel in accelerators]
                        )
                    ]
                )
            )
            # "N/A"
            # if resource["gpus"] is None
            # else [gpu["name"].strip() for gpu in resource["gpus"]],
        ]
        for resource in resource_information
    ]

    table = tabulate(
        resource_data,
        headers=[
            "ID",
            "Pipelines",
            "Current run",
            "Run queue",
            "Accelerators",
        ],
        tablefmt="psql",
    )

    print(table)



import json
import typing as t

from httpx import Response


def handle_stream_response(response: Response) -> t.Generator[t.Any, t.Any, None]:
    """Helper function to help with parsing streamed data from an API into
    JSON objects
    """
    incomplete_chunk = ""
    for chunk in response.iter_text():
        # chunk of data coming back from API could either be an incomplete
        # result or more than one results, so try to handle both these cases
        full_chunk = incomplete_chunk + chunk
        results = full_chunk.split("\n")
        for result in results:
            if not result:
                continue
            try:
                result_json = json.loads(result)
                incomplete_chunk = ""
            except json.JSONDecodeError:
                # assume due to an incomplete chunk of JSON
                incomplete_chunk = result
                break

            yield result_json


async def handle_async_stream_response(
    response: Response,
) -> t.AsyncGenerator[t.Any, t.Any]:
    """Helper function to help with parsing streamed data from an API into
    JSON objects

    Same as above but for async httpx client.
    """
    incomplete_chunk = ""
    async for chunk in response.aiter_text():
        # chunk of data coming back from API could either be an incomplete
        # result or more than one results, so try to handle both these cases
        full_chunk = incomplete_chunk + chunk
        results = full_chunk.split("\n")
        for result in results:
            if not result:
                continue
            try:
                result_json = json.loads(result)
                incomplete_chunk = ""
            except json.JSONDecodeError:
                # assume due to an incomplete chunk of JSON
                incomplete_chunk = result
                break

            yield result_json



import json
import urllib.parse
from typing import Generator

from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from websockets.sync.client import connect

from pipeline import current_configuration
from pipeline.cloud import http
from pipeline.util.logging import _print


def get_run_logs(run_id: str) -> list[str] | None:
    response = http.get(
        f"/v4/logs/run/{run_id}",
    )
    response_json = response.json()
    return response_json.get("log_entries")


def tail_run_logs(run_id: str) -> Generator[tuple, None, None]:
    url_info = urllib.parse.urlparse(current_configuration.active_remote.url)
    target_host = url_info.hostname
    target_port = url_info.port
    if target_port is None:
        target_port = 80 if url_info.scheme == "http" else 443

    protocol = "ws" if url_info.scheme == "http" else "wss"

    connection_string = urllib.parse.urljoin(
        protocol + "://" + str(target_host) + ":" + str(target_port),
        f"/v3/logs/run/{run_id}?follow=true",
    )
    try:
        with connect(
            connection_string,
            additional_headers={
                "Authorization": f"Bearer {current_configuration.active_remote.token}"
            },
        ) as websocket:
            message = websocket.recv()
            while True:
                stream_data: dict = json.loads(json.loads(message))
                streams = stream_data.get("streams", None)
                if streams is None:
                    raise Exception("No streams in message")

                if len(streams) == 0:
                    raise Exception("No streams in message")

                for stream in streams:
                    values = stream.get("values", None)
                    if values is None:
                        raise Exception("No values in stream")
                    for value in values:
                        yield value
                try:
                    message = websocket.recv()
                except ConnectionClosedOK:
                    return
    except ConnectionClosedOK as e:
        if e.code == 1000:
            _print("Log stream disconnected", level="WARNING")
        else:
            _print(
                f"Log stream disconnected with code {e.code}, reason {e.reason}",
                level="ERROR",
            )
    except ConnectionClosedError as e:
        if e.code == 4000:
            _print(f"Run {run_id} not found", level="ERROR")


def get_pipeline_startup_logs(pipeline_id_or_pointer: str) -> list[str] | None:
    response = http.get(
        f"/v4/logs/pipeline-startup/{pipeline_id_or_pointer}",
    )
    response_json = response.json()
    return response_json.get("log_entries")



import numpy as np
import torch

from pipeline.objects import pipe


@pipe
def tensor_to_list(input_tensor: torch.Tensor) -> list:
    return np.array(input_tensor.cpu().detach().numpy(), dtype=np.float64).tolist()



from datetime import datetime

from pipeline.cloud.schemas import BaseModel


class RemoteFileData(BaseModel):
    id: str
    path: str
    url: str | None


class FileGet(BaseModel):
    id: str
    path: str

    created_at: datetime
    updated_at: datetime


class MultipartFileUploadInitCreate(BaseModel):
    name: str | None = None


class MultipartFileUploadInitGet(BaseModel):
    file_id: str
    upload_id: str


class MultipartFileUploadPartCreate(BaseModel):
    file_id: str
    upload_id: str
    # The part number for this multi-part file upload
    part_num: int


class MultipartFileUploadPartGet(BaseModel):
    # The URL to use when uploading the file chunk
    upload_url: str


class MultipartFileUploadMetadata(BaseModel):
    """Schema for multi-part uploads direct to storage server.

    (We don't have any control over this schema)
    """

    ETag: str
    PartNumber: int


class MultipartFileUploadFinaliseCreate(BaseModel):
    file_id: str
    upload_id: str
    # The metadata obtained from each part of the file upload
    multipart_metadata: list[MultipartFileUploadMetadata]


class UploadFileUsingPresignedUrl(BaseModel):
    local_file_path: str
    upload_url: str
    upload_fields: dict[str, str]


class UploadFilesToRemoteStorageCreate(BaseModel):
    files: list[UploadFileUsingPresignedUrl]



import json
from argparse import ArgumentParser, Namespace, _SubParsersAction
from datetime import datetime
from enum import Enum

from tabulate import tabulate

from pipeline.cloud import http
from pipeline.cloud.schemas.pagination import (
    Paginated,
    get_default_pagination,
    to_page_position,
)
from pipeline.util.logging import _print


class ScalingConfigType(str, Enum):
    windows = "windows"


def _create_scaling_config(namespace: Namespace) -> None:
    name = getattr(namespace, "name")
    type_ = getattr(namespace, "type")
    args_ = getattr(namespace, "args")
    min_nodes = getattr(namespace, "min_nodes")
    max_nodes = getattr(namespace, "max_nodes")
    # Annoyingly, setting default values above did not seem to work
    payload = {}
    payload["name"] = name
    payload["type"] = type_ or "windows"
    payload["args"] = args_ or {}
    payload["minimum_nodes"] = min_nodes or 1
    payload["maximum_nodes"] = max_nodes or 100

    result = http.post("/v4/scaling-configs", json_data=payload)

    scaling_config = result.json()["name"]

    _print(f"Created scaling configuration {scaling_config}")


def _get_scaling_config(args: Namespace) -> None:
    _print("Getting scaling configurations")

    params = dict()
    pagination = get_default_pagination()
    if name := getattr(args, "name", None):
        params["name"] = name
    if skip := getattr(args, "skip", None):
        pagination.skip = skip
    if limit := getattr(args, "limit", None):
        pagination.limit = limit
    paginated_scaling_configs: Paginated[dict] = http.get(
        "/v4/scaling-configs",
        params=dict(**params, **pagination.dict()),
    ).json()

    scaling_configs = [
        [
            scaling["id"],
            scaling["name"],
            datetime.fromtimestamp(scaling.get("created_at"))
            if "created_at" in scaling
            else "N/A",
            scaling["type"],
            scaling["args"],
        ]
        for scaling in paginated_scaling_configs["data"]
    ]

    page_position = to_page_position(
        paginated_scaling_configs["skip"],
        paginated_scaling_configs["limit"],
        paginated_scaling_configs["total"],
    )

    table = tabulate(
        scaling_configs,
        headers=[
            "ID",
            "Name",
            "Created",
            "Type",
            "Args",
        ],
        tablefmt="psql",
    )
    print(table)
    print(f"\nPage {page_position['current']} of {page_position['total']}\n")


def _edit_scaling_config(args: Namespace) -> None:
    # If you add new arg, check whether it should be add to
    # the `if all(...)` below
    name = getattr(args, "name")
    type_ = getattr(args, "type", None)
    args_ = getattr(args, "args", None)
    min_nodes = getattr(args, "min_nodes")
    max_nodes = getattr(args, "max_nodes")

    if all(arg is None for arg in (type_, args_, min_nodes, max_nodes)):
        _print("Nothing to edit.", level="ERROR")
        return

    payload = {}
    if type_ is not None:
        payload["type"] = type_
    if args_ is not None:
        payload["args"] = args_
    if min_nodes is not None:
        payload["minimum_nodes"] = min_nodes
    if max_nodes is not None:
        payload["maximum_nodes"] = max_nodes

    response = http.patch(
        f"/v4/scaling-configs/{name}",
        payload,
    )
    _print(response.json())

    _print("Scaling configuration edited!")


def _delete_scaling_config(args: Namespace) -> None:
    name = getattr(args, "name")

    http.delete(
        f"/v4/scaling-configs/{name}",
    )

    _print("Scaling configuration deleted!")


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "scalings",
        aliases=["scaling"],
        help="Get scaling config information.",
    )

    get_parser.set_defaults(func=_get_scaling_config)

    # get by name
    get_parser.add_argument(
        "--name",
        "-n",
        help="Scaling config name.",
        type=str,
    )
    get_parser.add_argument(
        "--skip",
        "-s",
        help="Number of scaling configs to skip in paginated set.",
        type=int,
    )
    get_parser.add_argument(
        "--limit",
        "-l",
        help="Total number of scaling configs to fetch in paginated set.",
        type=int,
    )


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    create_parser = command_parser.add_parser(
        "scalings",
        aliases=["scaling"],
        help="Create a new scaling configuration.",
    )

    create_parser.set_defaults(func=_create_scaling_config)

    create_parser.add_argument(
        "name",
        type=str,
        help="Scaling configuration to create.",
    )
    create_parser.add_argument(
        "--type",
        "-t",
        help="The type of the scaling configuration",
        type=ScalingConfigType,
    )
    create_parser.add_argument(
        "--args",
        "-a",
        help="The arguments of the scaling configuration",
        type=json.loads,
    )
    create_parser.add_argument(
        "--min-nodes",
        help="The minimum number of nodes for the scaling configuration",
        type=int,
    )
    create_parser.add_argument(
        "--max-nodes",
        help="The maximum number of nodes for the scaling configuration",
        type=int,
    )


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    edit_parser = command_parser.add_parser(
        "scalings",
        aliases=["scaling"],
        help="Edit scaling config information.",
    )

    edit_parser.set_defaults(func=_edit_scaling_config)

    # Requires name param to edit
    edit_parser.add_argument(
        "name",
        help="The name of the scaling configuration to edit.",
        type=str,
    )
    edit_parser.add_argument(
        "--type",
        "-t",
        help="The type of the scaling configuration",
        type=ScalingConfigType,
    )
    edit_parser.add_argument(
        "--args",
        "-a",
        help="The arguments of the scaling configuration",
        type=json.loads,
    )
    edit_parser.add_argument(
        "--min-nodes",
        help="The minimum number of nodes of the scaling configuration",
        type=int,
    )
    edit_parser.add_argument(
        "--max-nodes",
        help="The maximum number of nodes of the scaling configuration",
        type=int,
    )


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    delete_parser = command_parser.add_parser(
        "scalings",
        aliases=["scaling"],
        help="Delete a scaling configuration.",
    )

    delete_parser.set_defaults(func=_delete_scaling_config)

    delete_parser.add_argument(
        "name",
        help="Name of the scaling configuration to delete.",
    )
